{$A+,B-,F+,G+,N+,P+,Q-,R-,T-,V+,X+}
{****************************************************************************
 * S3 driver for GrafX2 - copyright (c) 1998-1999 by Stefan Goehler         *
 * Currently only Trio chipsets supported, still Vesa-based for             *
 * initialization :(                                                        *
 * References: WhatVGA sources and documentation, GrafX 1.x S3 driver       *
 * Known bugs: .doesn't work on newer S3 cards (beginning with Virge)       *
 *             .setvirtualscreenwidth doesn't work in 32 bpp with demo      *
 *              program                                                     *
 ****************************************************************************}
{$IFNDEF DPMI}
  !Real mode not supported!
  This unit is fully 32 bit oriented, so forget about real mode.
  You need BP7 to use protected mode. If you have it, please
  switch at compiler/platform to PM (start bp.exe!)
{$ENDIF}
unit drv_s3;
interface
uses gx_glob,winapi,newfrontier,drv_pci,ports;

implementation
type
  tvesainfo = record
    VESASignature : array[0..3] of char;
    VESAVersion   : word;
    OEMStringPtr  : pchar;
    Capabilities  : longint;
    VideoModePtr  : pointer;
    totalmemory   : word;
    { beware - this will only be filled if VBE 2+ is available }
    OemSoftwareRev    : word;
    OemVendorNamePtr  : pchar;
    OemProductNamePtr : pchar;
    OemProductRevPtr  : pchar;
    reserved          : array[1..222] of byte;
  end;

  tmodeinfo = record
    ModeAttributes      : word;
    WinAAttributes      : byte;
    WinBAttributes      : byte;
    WinGranularity      : word;
    Winsize             : word;
    WinASegment         : word;
    WinBSegment         : word;
    WinFuncPtr          : pointer;
    BytesPerScanLine    : word;
    XResolution         : word;
    YResolution         : word;
    XCharsize           : byte;
    YCharsize           : byte;
    NumberOfPlanes      : byte;
    BitsPerPixel        : byte;
    NumberOfBanks       : byte;
    MemoryModel         : byte;
    BankSize            : byte;
    NumberOfImagePages  : byte;
    Reserved1           : byte;
    RedMaskSize         : byte;
    RedFieldPosition    : byte;
    GreenMaskSize       : byte;
    GreenFieldPosition  : byte;
    BlueMaskSize        : byte;
    BlueFieldPosition   : byte;
    RsvdMaskSize        : byte;
    RsvdFieldPosition   : byte;
    DirectColorModeInfo : byte;
    { VBE 2 Variables }
    PhysBasePtr         : longint;
    OffScreenMemOffset  : longint;
    OffScreenMemSize    : word;
    { VBE 3 Variables }
    LinBytesPerScanLine   : longint;
    BnkNumberOfImagePages : byte;
    LinNumberOfImagePages : byte;
    LinRedMaskSize        : byte;
    LinRedFieldPosition   : byte;
    LinGreenMaskSize      : byte;
    LinGreenFieldPosition : byte;
    LinBlueMaskSize       : byte;
    LinBlueFieldPosition  : byte;
    LinRsvdMaskSize       : byte;
    LinRsvdFieldPosition  : byte;
    MaxPixelClock         : longint;
    Reserved              : array[1..188] of byte;
  end;


var
  vesainfo : tvesainfo;
  modeinfo : tmodeinfo;
  selector : longint;
  modeptr  : pmode;
  fastio   : boolean;
  hwasupp  : boolean;

function intr32(intno : byte;var regs : trealregs) : boolean; assembler;
asm
  mov   ax,300h
  sub   bx,bx
  mov   bl,intno
  sub   cx,cx
  les   di,regs
  int   31h
  cmc
  mov   ax,0
  adc   ax,ax
end;

function getvesainfo(var vesainfo : tvesainfo) : boolean;
var
  regs : trealregs;
begin
  getvesainfo := false;
  selector := globaldosalloc(512);
  fillbyte(farptr(@regs),sizeof(trealregs),0);
  fillchar(regs,sizeof(regs),0);
  move2(farptr(@vesainfo),ptr48(selector,0),sizeof(vesainfo));
  with regs do begin
    reales  := selector shr 16;
    realedi := 0;
    realeax := $4F00;
  end;
  intr32($10,regs);
  move2(ptr48(selector,0),farptr(@vesainfo),sizeof(vesainfo));
  globaldosfree(selector);
  regs.realeax := regs.realeax and $FFFF;
  getvesainfo := regs.realeax = $4F;
end;


function getmodeinfo(var modeinfo : tmodeinfo;mode : word) : boolean;
var
  regs : trealregs;
begin
  selector := globaldosalloc(sizeof(tmodeinfo));
  fillbyte(ptr48(selector,0),sizeof(tmodeinfo),0);
  fillchar(regs,sizeof(regs),0);
  with regs do begin
    reales  := selector shr 16;
    realedi := 0;
    realeax := $4F01;
    realecx := mode;
  end;
  intr32($10,regs);
  move2(ptr48(selector,0),farptr(@modeinfo),sizeof(tmodeinfo));
  globaldosfree(selector);
  regs.realeax := regs.realeax and $FFFF;
  getmodeinfo := regs.realeax = $4F;
end;


procedure fillmodetable_vesa(var table);
type
  tvideomode = array[0..511] of word;
const
  supportedbpp : set of byte=[{8,}15,16,24,32];
var
  i,rtab    : word;
  videomode : ^tvideomode;
begin
  i          := 0;
  getmem(videomode,1024);
  move(ptr(getrmselector(longint(vesainfo.videomodeptr) shr 16),longint(vesainfo.videomodeptr) and $FFFF)^,
       videomode^,1024);
  for rtab := 0 to 4 do tmodetable(table)[rtab].avmodes := 0;
  while (videomode^[i] <> $FFFF)and(i < 512) do begin
    if getmodeinfo(modeinfo,videomode^[i]) then
    if (modeinfo.memorymodel <> 0)and(modeinfo.bitsperpixel in supportedbpp)
    then begin
      case modeinfo.bitsperpixel of
        8  : rtab := 0;
        15,16 : rtab := modeinfo.greenmasksize-4;
        24 : rtab := 3;
        32 : rtab := 4;
        else rtab := 5;
      end;
      if (rtab < 5)and(tmodetable(table)[rtab].avmodes < 25)and(
         longint(modeinfo.bytesperscanline)*modeinfo.yresolution <=
         longint(vesainfo.totalmemory) shl 16)and
         (getbit(modeinfo.modeattributes,4))and
         (getbit(modeinfo.modeattributes,0)) then begin
        tmodetable(table)[rtab].mode[tmodetable(table)[rtab].avmodes].resx := modeinfo.xresolution;
        tmodetable(table)[rtab].mode[tmodetable(table)[rtab].avmodes].resy := modeinfo.yresolution;
        tmodetable(table)[rtab].mode[tmodetable(table)[rtab].avmodes].mode := videomode^[i];
        inc(tmodetable(table)[rtab].avmodes);
      end;
    end;
    inc(i);
  end;
  freemem(videomode,1024);
end;

procedure decbank_s3_new;far;assembler;
{ decrease bank of newer s3 cards }
asm
  push  ax
  push  dx
  mov   ah,byte ptr lastbank
  dec   ah
  mov   dx,3D4h
  mov   al,6Ah
  out   dx,ax
  mov   byte ptr lastbank,ah
  mov   byte ptr lastbank[1],0
  pop   dx
  pop   ax
end;

procedure decbank_s3_std;far;assembler;
{ decrease bank of older s3 cards }
asm
  pusha
  mov   dx,3d4h
  mov   al,35h
  mov   ah,byte ptr lastbank
  dec   ah
  mov   byte ptr lastbank,ah
  mov   byte ptr lastbank[1],0
  mov   bl,ah
  and   ah,$F
  out   dx,ax

  mov   al,51h
  out   dx,al
  inc   dx
  in    al,dx
  and   al,243
  shr   bl,2
  and   bl,1100b
  or    al,bl
  out   dx,al
  popa
end;

procedure incbank_s3_new;far;assembler;
{ increase bank of newer s3 cards }
asm
  push  ax
  push  dx
  mov   ah,byte ptr lastbank
  inc   ah
  mov   dx,3D4h
  mov   al,6Ah
  out   dx,ax
  mov   byte ptr lastbank,ah
  pop   dx
  pop   ax
end;

procedure incbank_s3_std;far;assembler;
{ increase bank of older s3 cards }
asm
  pusha
  mov   dx,3d4h
  mov   al,35h
  mov   ah,byte ptr lastbank
  inc   ah
  mov   byte ptr lastbank,ah
  mov   byte ptr lastbank[1],0
  mov   bl,ah
  and   ah,$F
  out   dx,ax

  mov   al,51h
  out   dx,al
  inc   dx
  in    al,dx
  and   al,243
  shr   bl,2
  and   bl,1100b
  or    al,bl
  out   dx,al

  popa
end;

procedure setbank_s3_new(bank : word);far;assembler;
{ set bank of newer s3 cards }
asm
  mov   ah,byte ptr bank
  cmp   ah,byte ptr lastbank
  je    @end
    mov   dx,3D4h
    mov   al,6Ah
    out   dx,ax
    mov   byte ptr lastbank,ah
  @end:
end;

procedure setbank_s3_std(bank : word);far;assembler;
{ set bank of older s3 cards }
asm
  mov   ax,bank
  cmp   ax,lastbank
  je    @end

  mov   lastbank,ax
  mov   ah,al
  mov   dx,3d4h
  mov   al,35h
  mov   bl,ah
  and   ah,$F
  out   dx,ax

  mov   al,51h
  out   dx,al
  inc   dx
  in    al,dx
  and   al,243
  shr   bl,2
  and   bl,1100b
  or    al,bl
  out   dx,al

  @end:
end;


procedure unlocks3regs;assembler;
asm
  mov   dx,3D4h
  mov   al,38h
  out   dx,al
  inc   dx
  mov   al,48h
  out   dx,al
  dec   dx

  mov   al,39h
  out   dx,al
  inc   dx
  mov   al,0A5h
  out   dx,al
  dec   dx

  mov   al,40h
  out   dx,al
  inc   dx
  in    al,dx
  or    al,1
  out   dx,al
end;

procedure locks3regs;assembler;
asm
  mov   dx,3D4h
  mov   al,40h
  out   dx,al
  inc   dx
  in    al,dx
  and   al,0FEh
  out   dx,al
end;

const
  MISC_WT   = $03C2;
  MISC_RD   = $03CC;

  STATUS_0  = $03C2;
  STATUS_1  = $03DA;

  SEQ_ADR   = $03C4;            { Sequencer Registers: SR0-4 }
  SEQ_DATA  = $03C5;

  CRTC_ADR  = $03D4;            { Crt Controller Registers: CR0-18,22,24/26 }
  CRTC_DATA = $03D5;

  ATR_ADR   = $03C0;            { Attribute Controller Registers: AR0-14 }
  ATR_DATA  = $03C1;

  SR = SEQ_ADR;
  CR = CRTC_ADR;

function ReadReg(Address: Word; Index: Byte): Byte;
{ Read a VGA register. }
var
  Tmp: Byte;
begin
  if Address = ATR_ADR then begin
    Tmp := inp(STATUS_1);
    outp(ATR_ADR,(inp(ATR_ADR) and $E0) or (Index and $1F));
    ReadReg := inp(ATR_DATA);
  end { if }
  else begin
    outp(Address,Index);
    ReadReg := inp(Address+1);
  end; { else }
end; { ReadReg }

function WriteReg(Address: Word; Index, Value: Byte): Byte;
{ Write to a VGA register. }
var
  Tmp: Byte;
begin
  if Address <> ATR_ADR then begin
    outp(Address,Index);
    outp(Address+1,Value);
  end { if }
  else begin
    Tmp := inp(STATUS_1);
    outp(ATR_ADR,(inp(ATR_ADR) and $E0) or (Index and $1F));
    outp(ATR_ADR,Value);
  end; { else }
end; { WriteReg }

function RegBit(PAddr : word;mask : byte): Boolean;
var old : byte;
begin
  RegBit := true;
  old := port[PAddr];                                { save register    }
  port[PAddr] := old and not mask;                   { reset bits       }
  if port[PAddr] and mask <> 0 then RegBit := false
  else begin
    port[PAddr] := old or mask;                      { set bits         }
    if port[PAddr] and mask <> mask then RegBit := false;
  end;
  port[PAddr] := old;                                { restore register }
end;

procedure setdisplaypos_s3(x,y : word);
var
  a    : longint;
  temp : byte;
begin
  a := longint(y)*modeptr^.bpl+x*modeptr^.pixel;

  outp($3D4,$0D);
  outp($3D5,a shr 2);
  outp($3D4,$0C);
  outp($3D5,a shr 10);

  if not fastio then begin
    outp($3D4,$31);
    temp := inp($3D5);
    temp := temp and $CF;
    temp := temp or ((a shr 14) and not $CF);
    outp($3D5,temp);

    outp($3D4,$51);
    temp := inp($3D5);
    temp := temp and not 3;
    temp := temp or ((a shr 20) and 3);
    outp($3D5,temp);
  end else begin
    outp($3D4,$69);
    outp($3D5,(a shr 18) and $1F);
  end;
end;

function setvirtualscreenwidth_s3(width : word) : word;
var
  val : word;
begin
  width := (width+7) and $FFF0;
  val   := (width shr 3)*modeptr^.pixel;
  writereg(cr,$13,val);
  val   := ((val shr 8) and $03) shl 4;
  writereg(cr,$51,readreg(cr,51) and not $30 or val);
  setvirtualscreenwidth_s3 := width;
end;


procedure s3_bitblit(x1,y1,x2,y2,dx,dy : longint);
var
  dir   : word;
  sx,sy : longint;
begin
  if (y1 > y2) then swaplong(y1,y2);
  if (x1 > x2) then swaplong(x1,x2);

  sx := x2-x1;
  sy := y2-y1;

  if (dy < y1) or ((y1 = dy) and (dx < x1)) then dir := 0
  else begin
    dir := 1;
    x1  := x1+sx;
    y1  := y1+sy;
    dx  := dx+sx;
    dy  := dy+sy;
  end;

 (* if (pixels>1024) and (memmode>=_p8) and
    (cv.Version>=S3_928) and (cv.Version<S3_864) then  {Argh...}
  begin
    srcY:=srcY*2;
    dstY:=dstY*2;
    dY:=dY*2;
  end;*)
  repeat until inp($9AE8) = 0;
  outpw($BAE8,$67);
  outpw($BEE8,$A000);
  outpw($86E8,x1);
  outpw($82E8,y1);
  outpw($8EE8,dx);
  outpw($8AE8,dy);

  outpw($96E8,sx);
  outpw($BEE8,sy);
  repeat until inp($9AE8) = 0;
  if (dir <> 0) then outpw($9AE8,$C013) else outpw($9AE8,$C0F3);
  repeat until inpw($9AE8) and $200 = 0; { wait until engine finished }
end;

procedure s3_fillrect(x1,y1,sizex,sizey,col,flags : longint);
begin
(*  if (pixels>1024) and (memmode>=_p8) and
    (cv.Version>=S3_928) and (cv.Version<S3_864) then  {Argh...}
  begin
    yst:=yst*2;
    dy:=dy*2;
  end;*)
  repeat until inp($9AE8) = 0;
  outpw($86E8,x1);
  outpw($82E8,y1);
  outpw($96E8,sizex-1);
  outpw($BEE8,$E000);
  outpw($A6E8,col);
  if (modeptr^.bpp >= 24) then
  begin
    outpw($BEE8,$E010);
    outpw($A6E8,col shr 16);
  end;
  repeat until inp($9AE8) = 0;
  outpw($BEE8,$A000);
  case flags shr 24 of
    $00 : outpw($BAE8,$27); { normal filling }
    $04 : outpw($BAE8,$25); { xor filling    }
    else exit;
  end;
  outpw($BEE8,sizey);
  outpw($BEE8,$A000);
  outpw($9AE8,$40F1);
  repeat until inpw($9AE8) and $200 = 0; { wait until engine finished }
end;

procedure s3_clear;
begin
(*  if (pixels>1024) and (memmode>=_p8) and
    (cv.Version>=S3_928) and (cv.Version<S3_864) then  {Argh...}
  begin
    yst:=yst*2;
    dy:=dy*2;
  end;*)
  repeat until inp($9AE8) = 0;
  outpw($86E8,0);
  outpw($82E8,modeptr^.pageadd);
  outpw($96E8,modeptr^.maxx);
  outpw($BEE8,$E000);
  outpw($A6E8,0);
  if (modeptr^.bpp >= 24) then
  begin
    outpw($BEE8,$E010);
    outpw($A6E8,0);
  end;
  repeat until inp($9AE8) = 0;
  outpw($BEE8,$A000);
  outpw($BAE8,$27); { normal filling }
  outpw($BEE8,modeptr^.maxy);
  outpw($BEE8,$A000);
  outpw($9AE8,$40F1);
  repeat until inpw($9AE8) and $200 = 0; { wait until engine finished }
end;


function detect_s3 : word;forward;
procedure closevideo_s3(var mode : tmode);forward;
function setupvideo_s3(modenum : longint;var mode : tmode;flags : longint) : word;forward;

const
  s3driver : tgfxdriver=(
  info:
   (drvname       : 'S3 testal driver';
    drvversion    : $0020;
    drvsubversion : $0516;
    drvversionstr : '0.21.0516';
    drvprod       : 'Stefan Goehler/Crossfire';
    copyright     : '(c) 1999 by Stefan Goehler';
    producer      : 'S3';
    name          : '';
    rev           : 0;
    subrev        : 0;
    capabilities:
    (cinterface  : gdbanked+gdlinear+gdsetoffset+gdvwidth;
     cclipping   : 0;
     cbitmapping : 0;
     cprimitives : 0;
     cpolygons   : 0);
    memory        : 0;
    linearaddress : 0;
    bustype       : 0;
    priority      : 2);
  proc:
   (detect        : detect_s3;
    fillmodetable : fillmodetable_vesa;
    fillrefreshtable : nil;
    setupvideo       : setupvideo_s3;
    closevideo       : closevideo_s3;
    setclip          : nil;
    setrefresh       : nil;
    fillrect         : s3_fillrect;
    bitblit          : s3_bitblit;
    scale            : nil;
    fillpoly         : nil;
    clear            : s3_clear;
    setbank          : setbank_s3_std;
    incbank          : incbank_s3_std;
    decbank          : decbank_s3_std;
    setdisplaypos    : setdisplaypos_s3;
    setvirtualscreenwidth : setvirtualscreenwidth_s3;
    savestate        : nil;
    restorestate     : nil;
    switchdacto8     : nil));

function initvesamode(mode : word;lfb : boolean) : boolean;assembler;
asm
           mov   bx,mode
           cmp   lfb,true
           jne   @nolfb   { Shall we use LFB?                         }
           or    bx,4000h { Yep - set LFB bit                         }
  @nolfb:  mov   ax,4F02h
           int   10h
           xor   ax,004Fh { ah <> 0 if mode fail, al <> 4F if no vesa }
           jnz   @fail
           mov   ax,1     { true                                      }
           jmp   @end
  @fail:   xor   ax,ax    { false                                     }
  @end:
end;

function inits3mode(resx,resy,bpp : word;lfb : boolean) : boolean;
var
  resval,colval : byte;
  val           : word;
begin
  inits3mode := false;
  case resx of
     640 : resval := $40;
     800 : resval := $80;
    1024 : resval := $00;
    1280 : resval := $C0;
    1152 : resval := $01;
    1600 : resval := $81;
    else exit;
  end;

  case bpp of
    8     : colval := $00;
    15,16 : colval := $10;
    24    : colval := $20;
    32    : colval := $30;
    else exit;
  end;
{  writereg($3D4,$50,resval or colval);}
{  case resx of
    1024 : resval := $00;
     640 : resval := $01;
     800 : resval := $02;
    1280 : resval := $03;
    1152 : resval := $08;
    1600 : resval := $0C;
    else exit;
  end;

  case bpp of
    8     : colval := $00;
    15,16 : colval := $10;
    24,32 : colval := $30;
    else exit;
  end;}

  resx := (resx+7) and $FFF0;
  val   := (resx shr 3)*((bpp+7) shr 3);
  writereg(cr,$13,val);
  val   := ((val shr 8) and $03) shl 4;
  writereg(cr,$51,readreg(cr,51) and not $30 or val);

    writereg(cr,$31,$09);
    writereg(cr,$34,$10);
    writereg(cr,$3A,$15);
    writereg(cr,$3B,$9C);
    writereg(cr,$3C,0);
    writereg(cr,$50,resval or colval);
    writereg(cr,$54,$70);
    writereg(cr,$5D,0);
    writereg(cr,$5E,0);
    writereg(cr,$5C,0);
    writereg(cr,$60,$FF);
    { Only for Trio yet }
    case bpp of
      8  : writereg(cr,$67,$10);
      15 : writereg(cr,$67,$30);
      16 : writereg(cr,$67,$50);
      24 : writereg(cr,$67,$70);
      32 : writereg(cr,$67,$D0);
    end;
(*    writereg(cr,$31,$09);
    writereg(cr,$34,$10);
    writereg(cr,$3A,$15);
    writereg(cr,$3B,$9C);
    writereg(cr,$3C,0);
    writereg(cr,$40,$21);
    writereg(cr,$50,0);
    writereg(cr,$51,0);
    writereg(cr,$54,$70);
{    writereg(cr,$5C,0);}
    writereg(cr,$5D,0);
    writereg(cr,$5E,0);
    writereg(cr,$60,$FF);
    writereg(cr,$67,0);*)

{    if lfb then begin
      writereg(cr,$58,readreg(cr,$58) and $EC);
      writereg(cr,$59,s3driver.info.linearaddress shr 24);
      writereg(cr,$5A,s3driver.info.linearaddress shr 16 and $FF);
      writereg(cr,$58,(readreg(cr,$58) and $E0) or $1E);
    end;}
  inits3mode := true;
  readln;
  halt;
end;

procedure Read3D4;assembler;
asm
  mov   dx,3d4h
  out   dx,al
  mov   ah,al
  inc   dx
  in    al,dx
  xchg  ah,al
  dec   dx
end;

procedure Write3D4;assembler;
asm
  mov   dx,3d4h
  out   dx,al
  xchg  ah,al
  inc   dx
  out   dx,al
  xchg  ah,al
  dec   dx
end;

procedure enableext;assembler;
asm
  mov   bl,0ECh
  mov   bh,10h
  mov   cx,000Ah
  mov   al,058h { linear aperture options / bit 2       }
  call  Read3D4 { ah = read(3d4h, 58h)                  }
  and   ah,0ECh { clear bit 0,1 and 4                   }
  call  Write3D4{ write(3d4h, 58h, new AH)              }
  mov   al,59h  { bit 0-1: linear memory address bit 8-9}
  mov   ah,ch
  call  Write3D4{ write(3d4h, 59h, 00h)                 }
  inc   al      { 05Ah : bit 0-7 = linear mem address bit 0-7}
  mov   ah,cl   { in 64k units ???                      }
  call  Write3D4{ write(3d4h, 5Ah, 0Ah)                 }
  mov   al,58h  { linear aperture options               }
  call  Read3D4 { ah = read(3d4h, 58h)                  }
  and   ah,bl   { and ah,ECh -> clear bit 0,1 and 4     }
  or    ah,bh   { or ah,10h (OR or ah,00h) -> set bit 4 }
  or    ah,1 shl 2{enable read ahead cache}
  call  Write3D4{ write(3d4h, 58h, new AH)              }
  mov   al,31h
  call  read3d4
  or    ah,1+8  { Enable Base Address Offset + Access to more than 256 KB}
  call  write3d4
  mov   al,54h
  call  read3d4
  and   ah,not 7
  or    ah,3    {Read Ahead Cache Extra Prefetch Control to 4}
  call  write3d4
end;

function setupvideo_s3(modenum : longint;var mode : tmode;flags : longint) : word;
var
  i : word;
  getmode : function(resx,resy : word;bpp : word;var table) : word;
begin
  unlocks3regs;

  modeptr := @mode;
  setupvideo_s3 := denoerror;
  if (flags and mdbanked = 0) then mode.flags := mode.flags or mdlinear;
  mode.mode := modenum;
  if mode.mode = 0 then begin
    setupvideo_s3 := demdnotav;
    exit;
  end;
  getmodeinfo(modeinfo,mode.mode);
  mode.memsize   := s3driver.info.memory;

  asm mov ax,3; int 10h;end; { needed by some S3-cards }
  if not initvesamode(mode.mode,false) then begin
    setupvideo_s3 := defail;
    locks3regs;
    exit;
  end;

{  inits3mode(modeinfo.xresolution,modeinfo.yresolution,modeinfo.bitsperpixel,true);}

  writereg(cr,$58,readreg(cr,$58) and $EC);
  if ((mode.flags and mdlinear) = 0) then begin
    mode.writeptr := getrmselector($A000);
    mode.readptr  := mode.writeptr;
    lastbank  := 65535;
    enableext;
  end else begin
    writereg(cr,$59,s3driver.info.linearaddress shr 24);
    writereg(cr,$5A,s3driver.info.linearaddress shr 16 and $FF);
    writereg(cr,$58,(readreg(cr,$58) and $E0) or $1E);
    mode.linearptr := physicaladdressmapping(s3driver.info.linearaddress,mode.memsize);
    mode.writeptr  := createdescriptor(mode.linearptr,mode.memsize);
    mode.readptr   := mode.writeptr;
    i              := getsegmentaccessrights(mode.writeptr);
    setsegmentaccessrights(mode.writeptr,(i and $FF70) or $93);
  end;

  mode.bpp        := modeinfo.bitsperpixel;
  mode.pixel      := (mode.bpp+7) shr 3;
  mode.resx       := modeinfo.xresolution;
  mode.resy       := modeinfo.yresolution;
  mode.bpl        := mode.resx*mode.pixel;
  mode.requiredlines := 8;
  mode.requiredmem   := mode.requiredlines*mode.bpl;
  mode.lasty      := mode.memsize div mode.bpl-mode.requiredlines-1;
  mode.numofpages := mode.memsize div (longint(mode.bpl)*mode.resy);
  setupvideo_s3   := denoerror;
end;

procedure closevideo_s3(var mode : tmode);
begin
  if (mode.flags and mdlinear <> 0)and(mode.writeptr <> 0) then begin
    freedescriptor(mode.writeptr);
    freephysicaladdressmapping(mode.linearptr);
  end;
  asm mov ax,textscreen.mode; int 10h;end;
  locks3regs;
end;

function detects3 : boolean;
var
  tmp   : byte;
  b     : boolean;
begin
  b := false;
  port[$3D4] := $38;   { Index Lock Register               }
  tmp := port[$3D5];   { save register                     }
  port[$3D5] := $48;   { unlock extended S3-registers      }
  port[$3D4] := $35;   { Index Bank Select Register        }
  if RegBit($3D5,$0F) then begin { Bit 0..3 useable        }
    port[$3D4] := $38; { Index Lock Register               }
    port[$3D5] := 0;   { lock extended S3-registers        }
    port[$3D4] := $35; { Index Bank Select Register        }
    b := not regbit($3D5,$0F);
  end;                 { card is S3, if bits 0..3 are zero }
  port[$3D4] := $38;
  port[$3D5] := tmp;   { restore register                  }
  detects3 := b;
end;

function getchipname : string;
var
  s           : string[50];
begin
  fastio  := false;
  hwasupp := true;
  case word(s3driver.info.rev) of
      $81 : s := '911';
      $82 : s := '911A/924';
 $90..$95 : s := '928';
 $A0..$A8 : s := '801/805';
      $B0 : s := '928PCI';
  $C0,$C1 : begin;s := 'Vision 864';fastio := true;end;
  $D0,$D1 : begin;s := 'Vision 964';fastio := true;end;
      $5631 : begin;s := 'Virge 325';fastio := true;end;
      $8810 : begin;s := 'Trio32 732';fastio := true;end;
      $8811 : begin;
              if (s3driver.info.subrev and $40 <> 0) then s := 'Trio64V+ 765'
              else s := 'Trio64 764';fastio := true; end;
      $8812 : begin;s := 'Aurora64V+';fastio := true;hwasupp := false;end;
      $8814 : begin;s := 'Trio64UV+ 767';fastio := true;hwasupp := false;end;
      $8815 : begin;s := 'Aurora128';fastio := true;hwasupp := false;end;
$8802,$88B0,$88F0 : begin;s := 'Vision 968';fastio := true; end;

      $883D : begin;s := 'VirgeVX 988';fastio := true;hwasupp := false;end;
      $8880 : begin;s := 'Vision 866';fastio := true;end;
      $8890 : begin;s := 'Vision 868';fastio := true;end;
      $8901 : begin;s := 'Trio64V2 DX/GX 775/785';fastio := true;hwasupp := false;end;
      $8901 : begin
                fastio  := true;
                hwasupp := false;
                if (s3driver.info.subrev and 1 <> 0)
                then s := 'Trio64V2 GX 785' else s := 'Trio64V2 DX 775';
              end;
      $8902 : begin;s := 'Plato PX';fastio := true;hwasupp := false;end;
      $8904 : begin;s := 'Trio 3D';fastio := true;hwasupp := false;end;
      $8A01 : begin
                fastio  := true;
                hwasupp := false;
                if (s3driver.info.subrev and 1 <> 0)
                then s := 'Virge GX 385' else s := 'Virge DX 375';
              end;
      $8A10 : begin;s := 'Virge GX2 357';fastio := true;hwasupp := false;end;
      $8A20 : begin;s := 'Savage3D 391';fastio := true;hwasupp := false;end;
      $8C01 : begin;s := 'Virge MX 260';fastio := true;hwasupp := false;end;
      $8C03 : begin;s := 'Virge MX+ 280';fastio := true;hwasupp := false;end;
      else begin
        hwasupp := false;
        s := 'unknown (Rev:'+hexstr(s3driver.info.rev,8)+' Subrev:'+hexstr(s3driver.info.subrev,8)+')';
      end;
  end;
  getchipname := s;
  if hwasupp then with s3driver do begin
    with info.capabilities do begin
      cinterface  := cinterface or gdclear;
      cbitmapping := cbitmapping or gdblit;
      cprimitives := cprimitives or gdfrect;
    end;
  end;
  if fastio then begin
    s3driver.proc.setbank := setbank_s3_new;
    s3driver.proc.incbank := incbank_s3_new;
    s3driver.proc.decbank := decbank_s3_new;
  end;
end;

procedure getcardspecs;
var
  creg1  : byte;
  memory : longint;
begin
  asm
     mov   al,36h
     call  read3d4
     mov   creg1,ah
  end;
  memory := 0;
  case word(s3driver.info.rev) of
    $81,$82 :
          case (creg1 shr 5 and 1) of { 911/924 }
            0 : memory := 4; { 1M   }
            1 : memory := 2; { 512K }
          end;
    $90..$95,$B0 :
          case (creg1 shr 5 and 3) of { 988 }
            0 : memory := 8;  { 2M }
            1 : memory := 16; { 4M }
            2 : memory := 24; { 6M }
            3 : memory := 32; { 8M }
          end;
    $8A10 :
          case (creg1 shr 6 and 3) of { Virge GX2 357 }
            1 : memory := 16; { 4M }
            3 : memory := 8;  { 2M }
          end;
    else
    case (creg1 shr 5) of { All other chips }
      0 : memory := 16; { 4M    }
      2 : memory := 12; { 3M    }
      3 : memory := 32; { 8M    }
      4 : memory := 8;  { 2M    }
      5 : memory := 24; { 6M    }
      6 : memory := 4;  { 1M    }
      7 : memory := 2;  { 512 K }
    end;
  end;
  s3driver.info.memory := memory shl 18;

  case (creg1 and 3) of
    0 : s3driver.info.bustype := cdbuseisa;
    1 : s3driver.info.bustype := cdbusvl;
    2 : case word(s3driver.info.rev) of
          $81,$82 : s3driver.info.bustype := cdbusmca;
          $90..$95,$A0..$A8 : s3driver.info.bustype := cdbusisa;
          else s3driver.info.bustype := cdbuspci;
        end;
    3 : s3driver.info.bustype := cdbusisa;
  end;
end;

function detect_s3 : word;
var
  info : ^tpcidevinfo;
  i    : integer;
  num  : word;
begin
  detect_s3 := 0;
  if not detects3 then exit;
  if not getvesainfo(vesainfo) then exit;
  if (vesainfo.vesasignature = 'VESA')and(vesainfo.vesaversion >= $0102) then
  begin
    asm
      db 66h;xor   ax,ax
      mov   al,30h
      call  Read3D4
      mov   al,ah
      xor   ah,ah
      db 66h;mov   s3driver.info.rev.word,ax
      cmp   al,0E0h
      jnae  @next
        mov   al,2Dh
        call  read3d4
        mov   s3driver.info.rev.byte[1],ah
        mov   al,2Eh
        call  read3d4
        mov   s3driver.info.rev.byte[0],ah
        mov   al,2Fh
        mov   al,ah
        xor   ah,ah
        db 66h;mov   s3driver.info.subrev.word,ax
      @next:
    end;
    with s3driver.info do begin
      linearaddress := 0;
      if pciavailable then begin
        num := 0;
        for i := 0 to 8 do if pcifinddev($5333,i) <> 0 then
        begin
          new(info);
          pcigetdevinfo(pcifinddev($5333,i),info^);
          if info^.class = $300 then inc(num);
          dispose(info);
        end;
        detect_s3 := num;
        i := pcifinddev($5333,0);
        if i <> 0 then begin
          new(info);
          pcigetdevinfo(i,info^);
          linearaddress := info^.base0;
          dispose(info);
        end else begin
          detect_s3 := 0;
          exit;
        end;
      end else detect_s3 := 1;
      getcardspecs;
      name := getchipname;
      if linearaddress = 0 then linearaddress := $3E00000;
    end;
  end;
end;

begin
  registergfxdriver(@s3driver);
end.

