{$A+,B-,D-,E+,F-,G+,I-,L-,N+,O-,P+,Q-,R-,S-,T-,V+,X+,Y-}
unit fliclib2;
(**** FLCLIB ****************************************************************
 * Author    : Stefan Goehler, Max Maischein                                *
 * Version   : 2.0b                                                         *
 * Task      : flX-videoplayer, able to play multiple files at once         *
 * Copyright : See grafx2.txt; illegal copy without this file               *
 ******** HISTORY ***********************************************************
 * Version 1.0 : Original by Max Maischein                                  *
 * Version 1.1 : *removed some bugs                                         *
 *               +support for Hires-videos added                            *
 *               *rewritten to OOP to get multiple file support (my very    *
 *                first try with OOP :)                                     *
 * Version 1.2 : +support for old FLI's - added 2 procedures                *
 *               *fixed bug of Max Maischein in Palette-procedures          *
 * Version 2.0 : *Conversion to GrafX2 interface                            *
 *               +supports now streaming                                    *
 *               *much more accurate timing                                 *
 ******** KNOWN PROBLEMS ****************************************************
 * .has problems with the displayment of some special FLCs, but unlike the  *
 *  program PV, the player doesn't crash every time                         *
 ****************************************************************************)
(*
  Q&D .FLC player

  You are allowed to modify the program to suit your needs, but in all
  events, this notice and the authors (my) name have to remain intact
  within this source code, and you have to credit me in any derivative
  work (=program etc.) that results from this file.

  corion@informatik.uni-frankfurt.de
  Max_Maischein@fido.spam.de
*)
interface
uses flc,grafx2,gx_glob,timebase,newfrontier,stream32;


const
  readbuffsize : word = 65528;

type

  pflcplayer =^tflcplayer;
  tflcplayer = object
    constructor open(var stream : tstream32;var image : timage);
    destructor  done;
    function    getversionstring : string;
    function    nextframe : boolean;
    function    seek(frame : longint) : longint;
    function    getcurrframe : longint;
    function    getlastframe  : longint;
    procedure   loadnextframe;
    private
    procedure fli_color64;
    procedure fli_color256;
    procedure loadnewdata;
    procedure fli_worddelta;
    procedure fli_worddelta_1624;
    procedure fli_bytedelta;
    procedure fli_copy;
    procedure fli_black;
    procedure fli_brun;
    procedure fli_brun_1624;
    private
    instream   : pstream32;
    NextBlockStart    : longint;
    NextSubBlockStart : longint;
    readbuffstop : word;
    otherchunk : procedure(chunktype : byte);
    linepos    : integer;

    dataptr    : ^byte;
    overlap    : word;
    H          : TFLIHeader;
    CurrBlock  : TFLCBlock;
    SubBlock   : TFLCSubBlock;
    TimeDiff   : longint;
    CurrFrame  : integer;
    RestartPos : longint;
    Buffer     : pointer;
    BufEnd     : ^byte;
    BufSize    : longint;
    bytesleft  : longint;
    display    : ^timage;
    pixel      : byte;
    timer      : tztimer;
  end;

  function info_flc(name : string;var image_info : timageinfo) : word;

implementation

const
  clock_scale  = 4659 div 70;

function info_flc(name : string;var image_info : timageinfo) : word;
var
  f : file;
  h : ^tfliheader;
  w : word;
begin
  info_flc := errfilenotfound;
  assign(f,name);
  {$I-}
  filemode := 64;
  reset(f,1);
  {$I+}
  if ioresult <> 0 then exit;
  new(h);
  blockread(f,h^,sizeof(tfliheader),w);
  if (h^.formatid <> $AF11)and(h^.formatid <> $AF12) then begin
    info_flc := errfalseformat;
    dispose(h);
    close(f);
    exit;
  end;
  if w < sizeof(tfliheader) then begin
    info_flc := errfileerror;
    dispose(h);
    close(f);
    exit;
  end;
  with image_info do begin
    sizex        := h^.width;
    sizey        := h^.height;
    bpp          := h^.depth;
    greyscale    := false;
    alphachannel := false;
    transparency := false;
    compressed   := true;
    comptype     := 'RLE/Delta';
    interlaced   := false;  { I don't know yet how to detect this }
    frames       := h^.numframes;
    if h^.formatid = $AF12 then speed := 1000000 div h^.speed
    else speed := 4608000 div (h^.speed*clock_scale);
    trans_color  := 0;
    memsize      := (h^.width*h^.height*h^.depth) shr 3;
    disksize     := filesize(f);
    supported    := (h^.depth = 8);
    version      := h^.formatid-$AF10;
  end;
  dispose(h);
  close(f);
  info_flc := errnoerror;
end;

procedure tflcplayer.fli_color256;
var
  startcolor  : integer;
  j,i         : integer;
  currpacket  : pflccolorpacket;
  colorcount  : integer;
  packetcount : integer;
  p           : pflccolortriplet;
begin
  startcolor := 0;
  currpacket := buffer;
  inc(word(currpacket),2);
  packetcount := word(buffer^);
  for j := 0 to packetcount-1 do begin
    with currpacket^ do begin
      p := @data;
      inc(startcolor,skipcount);
      colorcount := colortriplets;
      if colorcount = 0 then colorcount := 256;
      for i := 0 to colorcount-1 do begin
        writemem32b(ptr48(display^.data,display^.palstart+(startcolor+i) shl 2),p^.b);
        writemem32b(ptr48(display^.data,display^.palstart+(startcolor+i) shl 2+1),p^.g);
        writemem32b(ptr48(display^.data,display^.palstart+(startcolor+i) shl 2+2),p^.r);
        inc(p);
      end;
      inc(startcolor,colorcount);
      currpacket := @p^;
    end;
  end;
end;

procedure tflcplayer.fli_color64;
var
  startcolor  : integer;
  j,i         : integer;
  currpacket  : pflccolorpacket;
  colorcount  : integer;
  packetcount : integer;
  p           : pflccolortriplet;
begin
  startcolor := 0;
  currpacket := buffer;
  inc(word(currpacket),2);
  packetcount := word(buffer^);
  for j := 0 to packetcount-1 do begin
    with currpacket^ do begin
      p := @data;
      inc(startcolor,skipcount);
      colorcount := colortriplets;
      if colorcount = 0 then colorcount := 256;
      for i := 0 to colorcount-1 do begin
        writemem32b(ptr48(display^.data,display^.palstart+(startcolor+i) shl 2),p^.b shl 2);
        writemem32b(ptr48(display^.data,display^.palstart+(startcolor+i) shl 2+1),p^.g shl 2);
        writemem32b(ptr48(display^.data,display^.palstart+(startcolor+i) shl 2+2),p^.r shl 2);
        inc(p);
      end;
      inc(startcolor,colorcount);
      currpacket := @p^;
    end;
  end;
end;


procedure tflcplayer.loadnewdata;
begin
  overlap := readbuffsize-(word(dataptr)-ofs(buffer^));
  move2(ptr48(seg(buffer^),word(dataptr)),farptr(buffer),overlap);
  if bytesleft-overlap > readbuffsize then begin
    instream^.read32(ptr48(seg(buffer^),ofs(buffer^)+overlap),readbuffsize-overlap);
    dec(bytesleft,readbuffsize-overlap);
  end else begin
    instream^.read32(ptr48(seg(buffer^),ofs(buffer^)+overlap),bytesleft);
    bytesleft := 0;
  end;
  dataptr := buffer;
end;

procedure tflcplayer.fli_brun;
var
  i,j     : integer;
  count   : byte;
  imgpos  : longint;
begin
  dataptr    := buffer;
  imgpos     := display^.imagestart;
  for i := 0 to h.height-1 do begin
    inc(dataptr);
    if (bytesleft > 0)and(word(dataptr)-ofs(buffer^) >= readbuffstop) then loadnewdata;
    linepos := 0;
    while linepos < h.width do begin
      count := dataptr^;
      inc(dataptr);
      if count and $80 <> 0 then begin
        asm neg [count] end;
        move2(farptr(dataptr),ptr48(display^.data,imgpos+linepos),count);
        inc(dataptr,count);
      end else begin
        fillbyte(ptr48(display^.data,imgpos+linepos),count,dataptr^);
        inc(dataptr);
      end;
      inc(linepos,count);
    end;
    inc(imgpos,display^.bpl);
  end;
end;

procedure tflcplayer.fli_brun_1624;
var
  i,j     : integer;
  count   : byte;
  imgpos  : longint;
begin
  dataptr    := buffer;
  imgpos     := display^.imagestart;
  for i := 0 to h.height-1 do begin
    inc(dataptr);
    if (bytesleft > 0)and(word(dataptr)-ofs(buffer^) >= readbuffstop) then loadnewdata;
    linepos := 0;
    while linepos < h.width do begin
      count := dataptr^;
      inc(dataptr);
      if count and $80 <> 0 then begin
        asm neg [count] end;
        move2(farptr(dataptr),ptr48(display^.data,imgpos+linepos),count*pixel);
        inc(dataptr,count*pixel);
      end else begin
        if h.depth = 24 then
        fill24(ptr48(display^.data,imgpos+linepos),count*3,longint(pointer(dataptr)^))
        else fillword(ptr48(display^.data,imgpos+linepos),count shl 1,dataptr^);
        inc(dataptr,pixel);
      end;
      inc(linepos,count*pixel);
    end;
    inc(imgpos,display^.bpl);
  end;
end;

procedure tflcplayer.fli_worddelta;
var
  numlines   : word;
  numpackets : integer;
  currline   : integer;
  i          : integer;
  xpos       : word;
  count      : byte;
  imgpos     : longint;
begin
  currline := 0;
  numlines := word(buffer^);
  dataptr  := buffer;
  inc(dataptr,2);
  imgpos   := display^.imagestart;
  while numlines > 0 do begin
    numpackets := word(pointer(dataptr)^);
    inc(dataptr,2);
    if (bytesleft > 0)and(word(dataptr)-ofs(buffer^) >= readbuffstop) then loadnewdata;
    case numpackets shr 14 of
      0 : begin
            xpos := 0;
            for i := 1 to numpackets do begin
              inc(xpos,dataptr^);inc(dataptr);
              count := dataptr^;inc(dataptr);
              if count and $80 = 0 then begin
                if currline < h.height then
                move2(farptr(dataptr),ptr48(display^.data,imgpos+xpos),count shl 1);
                inc(dataptr,count shl 1);
              end else begin
                asm neg [count] end;
                if currline < h.height then
                fillword(ptr48(display^.data,imgpos+xpos),count shl 1,word(pointer(dataptr)^));
                inc(dataptr,2);
              end;
              inc(xpos,count shl 1);
            end;
            inc(currline);
            inc(imgpos,display^.bpl);
            dec(numlines);
          end;
      2 : begin
            if currline < h.height then
            writemem32b(ptr48(display^.data,imgpos+h.width),byte(@numpackets));
            xpos := h.width;
            inc(currline);
            inc(imgpos,display^.bpl);
          end;
      3 : begin
            inc(imgpos,longint(abs(numpackets))*display^.bpl);
            inc(currline,abs(numpackets));
          end;
    end;
  end;
end;

procedure tflcplayer.fli_worddelta_1624;
var
  numlines   : word;
  numpackets : integer;
  currline   : integer;
  i          : integer;
  xpos       : word;
  count      : byte;
  imgpos     : longint;
begin
  currline := 0;
  numlines := word(buffer^);
  dataptr  := buffer;
  inc(dataptr,2);
  imgpos   := 0;
  while numlines > 0 do begin
    numpackets := word(pointer(dataptr)^);
    inc(dataptr,2);
    if (bytesleft > 0)and(word(dataptr)-ofs(buffer^) >= readbuffstop) then loadnewdata;
    case numpackets shr 14 of
      0 : begin
            xpos := 0;
            for i := 1 to numpackets do begin
              inc(xpos,word(dataptr^)*pixel);inc(dataptr);
              count := dataptr^;inc(dataptr);
              if count and $80 = 0 then begin
                if currline < h.height then
                move2(farptr(dataptr),ptr48(display^.data,imgpos+xpos),(word(count)*pixel) shl 1);
                inc(dataptr,(word(count)*pixel) shl 1);
              end else begin
                asm neg [count] end;
                if currline < h.height then begin
                  if h.depth = 24 then
                  fill24(ptr48(display^.data,imgpos+xpos),(word(count)*6),longint(pointer(dataptr)^)) else
                  filldword(ptr48(display^.data,imgpos+xpos),count shl 2,longint(pointer(dataptr)^));
                end;
                inc(dataptr,1+pixel);
              end;
              inc(xpos,(word(count)*pixel) shl 1);
            end;
            inc(currline);
            inc(imgpos,display^.bpl);
            dec(numlines);
          end;
      2 : begin
            writemem32b(ptr48(display^.data,imgpos+h.width*pixel),byte(@numpackets));
            xpos := h.width*pixel;
            inc(currline);
            inc(imgpos,display^.bpl);
          end;
      3 : begin
            inc(imgpos,longint(numpackets and $3FFF)*display^.bpl);
            inc(currline,numpackets and $3FFF);
          end;
    end;
  end;
end;

procedure tflcplayer.fli_bytedelta;
{ New one added by me to support old FLI's. Maybe a bit buggy yet }
var
  numlines   : word;
  numpackets : word;
  currline   : integer;
  i          : integer;
  xpos       : word;
  count      : byte;
  imgpos     : longint;
begin
  dataptr    := buffer;
  currline   := word(buffer^);inc(dataptr,2);
  numlines   := word(pointer(dataptr)^);inc(dataptr,2);
  imgpos     := display^.imagestart+longint(currline)*display^.bpl;
  while numlines > 0 do begin
    dec(numlines);
    numpackets := dataptr^;inc(dataptr);
    if (bytesleft > 0)and(word(dataptr)-ofs(buffer^) >= readbuffstop) then loadnewdata;
    xpos := 0;
    for i := 1 to numpackets do begin
      inc(xpos,dataptr^);inc(dataptr);
      count := dataptr^;inc(dataptr);
      if count and $80 = 0 then begin
        if currline < h.height then
        move2(farptr(dataptr),ptr48(display^.data,imgpos+xpos),count);
        inc(dataptr,count);
      end else begin
        asm neg [count] end;
        if currline < h.height then
        fillbyte(ptr48(display^.data,imgpos+xpos),count,dataptr^);
        inc(dataptr);
      end;
      inc(xpos,count);
    end;
    inc(currline);
    inc(imgpos,display^.bpl);
  end;
end;


procedure tflcplayer.fli_copy;
var
  y      : integer;
  imgpos : longint;
begin
  dataptr := buffer;
  imgpos  := display^.imagestart;
  for y := 0 to h.height-1 do begin
    inc(dataptr,h.width);
    if (bytesleft > 0)and(word(dataptr)-ofs(buffer^) >= readbuffstop) then loadnewdata;
    move2(farptr(dataptr),ptr48(display^.data,imgpos),h.width*pixel);
    inc(imgpos,display^.bpl);
  end;
end;

procedure tflcplayer.fli_black;
begin
  clearimage(display^,0);
end;

procedure otherchunk_norm(chunktype : byte);far;
begin
end;

constructor tflcplayer.open(var stream : tstream32;var image : timage);
begin
  if (stream.status <> stok) then fail;
  instream := @stream;
  otherchunk := otherchunk_norm;
  if (ioresult <> 0) then fail;
  instream^.read(h,sizeof(h));
  if (h.formatid <> $AF11)and(h.formatid <> $AF12) then begin
    instream^.done;
    fail;
  end;
  if (h.depth <> 8) then fail;
  if (image.id <> idimage)or(image.data = 0)or(image.bpp <> 8)or
     (image.flags and impalette = 0) then
  begin
    closeimage(image);
    openimage(image,h.width,h.height,8,1,impalette);
  end;
  display := @image;
  if memavail < readbuffsize then fail;
  getmem(buffer,readbuffsize);
  if h.formatid = $AF12 then begin
    instream^.seek(h.offsetframe1);
    timediff := h.speed*1000;
  end
  else TimeDiff := round(h.speed*1000000/70);
  currframe  := 0;
  pixel      := display^.pixel;
  restartpos := instream^.getpos;
  timer.init;
  timer.start;
  loadnextframe;
end;

destructor tflcplayer.done;
begin
  instream^.done;
  freemem(buffer,readbuffsize);
  timer.done;
end;

function tflcplayer.getversionstring : string;
begin
  getversionstring := 'FlicLib 2b.0502';
end;

function tflcplayer.nextframe : boolean;
var
  b : boolean;
begin
  b := timediff < timer.lap;
  if b then timer.start;
  nextframe := b;
end;

function tflcplayer.seek(frame : longint) : longint;
var
  i : integer;
begin
  if (frame < 0) then frame := 0;
  if (frame > h.numframes-1) then frame := h.numframes-1;
  seek := frame-1;
  if (frame = currframe) then exit;
  if (frame > currframe) then
  begin
    for i := 0 to frame-currframe do loadnextframe;
  end else
  begin
    currframe := 1;
    instream^.seek(restartpos);
    for i := 1 to frame do loadnextframe;
  end;
end;

function tflcplayer.getcurrframe : longint;
begin
  getcurrframe := currframe-1;
end;

function tflcplayer.getlastframe : longint;
begin
  getlastframe := h.numframes-1;
end;

procedure tflcplayer.loadnextframe;
var
  i : integer;
begin
  instream^.read(currblock,sizeof(currblock));
  nextblockstart := instream^.getpos-sizeof(currblock)+currblock.size;
  if currblock.size > 0 then
  for i := 1 to CurrBlock.SubBlocks do begin
    instream^.read(SubBlock,SizeOf(SubBlock));
    if ioresult <> 0 then exit;
    BufSize := SubBlock.Size - SizeOf(SubBlock);
    NextSubBlockStart := instream^.getpos + BufSize;
    BufEnd := Buffer;
    inc(BufEnd,BufSize+1);
    if bufsize > readbuffsize then begin
      bytesleft    := bufsize-readbuffsize;
      bufsize      := readbuffsize;
      readbuffstop := readbuffsize-(h.width shl 1);
    end else begin
      readbuffstop := readbuffsize;
      bytesleft := 0;
    end;
    instream^.read(buffer^,bufsize);
    if ioresult <> 0 then exit;
    case subblock.blockid of
      0  : ;
      4  : fli_color256;
      7  : fli_worddelta;
      11 : fli_color64;
      12 : fli_bytedelta;
      13 : fli_black;
      15 : fli_brun;
      16,26 : fli_copy;
{      27 : fli_worddelta_1624;}
      else otherchunk(subblock.blockid);
    end;
    instream^.seek(nextsubblockstart);
  end;
  instream^.seek(nextblockstart);
  inc(currframe);
  if currframe = h.numframes then begin
    currframe := 1;
    instream^.seek(restartpos);
  end;
end;


begin
end.