{$A+,B-,D+,E+,F+,G+,I-,L+,N+,P+,Q-,R-,S-,T-,V+,X+,Y+,F+}
{****************************************************************************
 * GrafX2 MouseLib - copyright (c) 1998-1999 by Stefan Goehler              *
 ****************************************************************************}

{
 Things to implement:
 -double-/triplebuffer support
 -interprete middle button as doubleclick if wished
 -PS/2 support
 -mouse autohide/autochange for areas
}
unit mouselib;

interface
uses gx_glob,grafx2,timebase,gx_col,dos,ports;

type
  tintevent = record
    buttons : word   ;
    relmove : tpoint ; { Relative movement since last position }
  end;

  tcursor = timage;

  pmouse = ^tmouse;
  tmouse = object
    pressed,                 { Currently pressed buttons              }
    up,                      { Recently released buttons              }
    down          : word;    { Recently clicked buttons               }
    event         : longint; { See evXXXX constants                   }
    background    : timage;  { Background of the cursor               }
    pos,lpos      : tpoint;  { Actual, last cursor position           }
    inited        : boolean; { Mouse type has been inited, is ready   }
    ondoubleclick : procedure(x,y,button : integer);
    onstartdrag   : procedure(x,y,button : integer);
    ondrag        : procedure(x,y,button : integer);
    onenddrag     : procedure(x,y,button : integer);
    onmouseup     : procedure(x,y,button : integer);
    onmousedown   : procedure(x,y,button : integer);
    onmousemove   : procedure(x,y,button : integer);
    onmouseany    : procedure(x,y,button : integer);
{    ondragdrop    : procedure(x,y,button : integer);
    ondragover    : procedure(x,y,button : integer);}
    procedure   resetprocs;  { Resets all above procedures to nil }
    procedure   hide;        { hides the cursor                   }
    procedure   show;        { shows the cursor                   }
    procedure   freeze;      { Stops the cursor movement (still moves internal) }
    procedure   unfreeze;    { Unfreezes cursor movement          }
    procedure   setpos(x,y : integer);
    procedure   setwindow(x1,y1,x2,y2 : longint);
    function    setcursor(var cr : tcursor) : word;
    function    there(x1,y1,x2,y2 : longint) : boolean;
    procedure   setlefthand(lefthand : boolean); { Set true for left-handers}
    function    getlefthand : boolean;
    procedure   setdoubleclick(time : word); { Set doubleclick time; def: 350 ms}
    function    getdoubleclick : word;
    constructor init(var obj : tgrafx;flags : word);
    destructor  done;
    private
    timer       : pztimer;
    timecount   : comp;
    firstshow   : boolean;
    hidecount   : integer;
    drawing     : boolean; { Is there currently any drawing action? }
    cursor      : tcursor; { Currently used cursor                  }
    visible     : boolean; { Cursor is visible                      }
    frozen      : boolean;
    bgpos       : tpoint;  { Last position of the cursor background }
    screen      : pgrafx;  { Where the cursor is drawn to           }
    mouse       : word;    { Used mouse                             }
    min,max     : tpoint;  { Minimum/maximum position               }
    dragstart   : tpoint;  { Pos where dragging started             }
    lefthander  : boolean;
    clicktime   : longint; { Clicking time for doubleclick in ms }
    procedure   handlemouseevent(var in_ev : tintevent);
  end;

  pmhregistered = ^tmhregistered;
  tmhregistered = record
                    obj  : pmouse;
                    flags: word;
                    next : pmhregistered;
                  end;

const
  mouselib_versionstr = 'MouseLib 1.0720';
  dummy : tmhregistered = (obj:nil;flags:0;next:nil);

  evnothing     = $0000;
  evmousemove   = $0001;
  evmousedrag   = $0002;
  evmousedouble = $0004;

  mbleft        = $0001;
  mbright       = $0002;
  mbmiddle      = $0004;

  dragsensitivy : tpoint  = (x:2;y:2);
var

  crdefault   : tcursor;
  crhourglass : tcursor;
  crhandpoint : tcursor;   { Pointing hand         }
{  crbeam      = 4; { Text selection cursor }


  function  globalinitmouse(mouse : word) : word;
  procedure globaldonemouse;

implementation

var
  globalevent : array[0..3] of tintevent;
  mice        : word;    { Max 4 mice supported }
  avbuttons   : array[0..3] of word;
  movecount   : array[0..3] of record
                  x,y : longint;
                end;


const
  RxR = 0;
  TxR = 0;
  IER = 1;
  MCR = 4;
  LCR = 3;
  LSR = 5;
  MSR = 6;
  DLL = 0;
  DLH = 1;

  N = 0;
  O = 8;
  E = 24;

  com_irqaddr : array[1..4] of word =(4,3,4,3);

  mt_drv = 1;
  mt_com = 2;
  mt_ps2 = 3;

var
  oldint           : array[1..4] of pointer;
  dataqueue        : array[1..4,0..3] of byte;
  queuepos         : array[1..4] of word;
  com_baseaddr     : array[1..4] of word;
  numofports       : word;
  mouse_type       : array[0..3] of word;
  mouse_list       : array[1..4] of word; { which com has got which mouse }
  mouse_com        : array[0..3] of word; { which mouse has got which com }
  searchmouse      : word;

procedure swapbuttons(var buttons : word);assembler;
asm
  les   di,buttons
  mov   ax,es:[di]
  mov   bx,ax
  mov   cx,ax
  and   bx,1
  shl   bx,1
  and   cx,2
  shr   cx,1
  or    bx,cx
  and   ax,0FFFCh
  or    ax,bx
  mov   es:[di],ax
end;


procedure drawcursor(var mouse : tmouse;pos : tpoint;var screen : tgrafx);
begin
  if mouse.drawing then exit;
  mouse.drawing := true;
  if not mouse.firstshow then
  screen.draw(mouse.bgpos.x-mouse.cursor.spotx,mouse.bgpos.y-mouse.cursor.spotx,
              mouse.background,drdefault);
  screen.get(pos.x-mouse.cursor.spotx,pos.y-mouse.cursor.spotx,
             pos.x+mouse.cursor.sizex-1-mouse.cursor.spotx,
             pos.y+mouse.cursor.sizey-1-mouse.cursor.spotx,
             mouse.background);
  screen.draw(pos.x-mouse.cursor.spotx,pos.y-mouse.cursor.spotx,
              mouse.cursor,drdefault);
              mouse.bgpos := mouse.pos;
  mouse.firstshow := false;
  mouse.drawing := false;
end;

constructor tmouse.init(var obj : tgrafx;flags : word);
var
  n : pmhregistered;
begin
  inited  := false;
  new(n);
  n^.next    := dummy.next;
  n^.obj     := @self;
  n^.flags   := flags;
  dummy.next := n;
  screen     := @obj;
  pos.x      := screen^.centerx;
  pos.y      := screen^.centery;
  bgpos      := pos;
  min.x      := 0;
  min.y      := 0;
  max.x      := screen^.maxx;
  max.y      := screen^.maxy;
  visible    := false;
  firstshow  := true;
  hidecount  := 1;
  if setcursor(crdefault) <> errnoerror then fail;
  new(timer,init);
  timer^.start;
  clicktime  := 350;
  mouse      := flags;
  inited     := true;
end;

destructor tmouse.done;
var
  a   : pmhregistered;
  tmp : pointer;
begin
  if not inited then exit;
  inited := false;
  a := @dummy;
  while(a^.next <> nil)and(a^.next^.obj <> @self) do a := a^.next;
  if a^.next <> nil then begin
    tmp     := a^.next^.next;
    closeimage(cursor);
    closeimage(background);
    dispose(timer,done);
    dispose(a^.next);
    a^.next := tmp;
  end;
end;

procedure tmouse.hide;
begin
  if not inited then exit;
  inc(hidecount);
  if hidecount = 1 then begin
    visible := false;
    drawing := true;
    if not firstshow then
    screen^.draw(bgpos.x-cursor.spotx,bgpos.y-cursor.spotx,
              background,drdefault);
    drawing := false;
  end;
end;

procedure tmouse.show;
begin
  if not inited then exit;
  if not inited then exit;
  dec(hidecount);
  if hidecount = 0 then begin
    firstshow := true;
    drawcursor(self,pos,screen^);
  end;
  visible := hidecount <= 0;
end;

procedure tmouse.freeze;
begin
  if not inited then exit;
  frozen := true;
end;

procedure tmouse.unfreeze;
begin
  if not inited then exit;
  frozen := false;
end;

procedure tmouse.setpos(x,y : integer);
begin
  if not inited then exit;
  hide;
  drawing := true;
  pos.x := x;
  pos.y := y;
  if pos.x < min.x then pos.x := min.x;
  if pos.x > max.x then pos.x := max.x;
  if pos.y < min.y then pos.y := min.y;
  if pos.y > max.y then pos.y := max.y;
  drawing := false;
  show;
end;

procedure tmouse.setwindow(x1,y1,x2,y2 : longint);
begin
  if not inited then exit;
  hide;
  drawing := true;
  min.x := x1;
  min.y := y1;
  max.x := x2;
  max.y := y2;
  if pos.x < min.x then pos.x := min.x;
  if pos.x > max.x then pos.x := max.x;
  if pos.y < min.y then pos.y := min.y;
  if pos.y > max.y then pos.y := max.y;
  drawing := false;
  show;
end;

function tmouse.setcursor(var cr : tcursor) : word;
var
  err : word;
begin
  hide;
  err := closeimage(cursor);
  setcursor := err;
  if err <> errnoerror then exit;
  if cr.flags and imalpha <> 0 then
    setcursor := duplicateimage(cr,cursor) else
    setcursor := convertimage(cr,cursor,screen^.mode.bpp);
  show;
end;

function tmouse.there(x1,y1,x2,y2 : longint) : boolean;
begin
  if not inited then exit;
  there := (pos.x >= x1)and(pos.y >= y1)and
                (pos.x <= x2)and(pos.y <= y2);
end;

procedure tmouse.setlefthand(lefthand : boolean);
begin
  lefthander := lefthand;
end;

function tmouse.getlefthand : boolean;
begin
  getlefthand := lefthander;
end;

procedure tmouse.setdoubleclick(time : word);
begin
  clicktime := time;
end;

function tmouse.getdoubleclick : word;
begin
  getdoubleclick := clicktime;
end;

procedure tmouse.resetprocs;
begin
  ondoubleclick := nil;
  onstartdrag   := nil;
  ondrag        := nil;
  onenddrag     := nil;
  onmouseup     := nil;
  onmousedown   := nil;
  onmousemove   := nil;
  onmouseany    := nil;
{  ondragdrop    := nil;
  ondragover    := nil;}
end;

procedure tmouse.handlemouseevent(var in_ev : tintevent);
var
  i,allmask : integer;
  bank,mask : word;
  ev        : tintevent;
begin
  ev := in_ev;
  if lefthander then swapbuttons(ev.buttons);
  if (screen^.mode.flags and mdlinear = 0) then bank := lastbank;
  up      := 0;
  down    := 0;
  allmask := 0;
  event   := 0;
  for i := 0 to avbuttons[mouse]-1 do begin
    mask := word(1) shl i;
    allmask := allmask or mask;
    if (ev.buttons and mask <> 0)and(pressed and mask = 0)
    then down := down or mask;
    if (ev.buttons and mask = 0)and(pressed and mask <> 0)
    then up := up or mask;
  end;
  pressed := ev.buttons and allmask;
  if down and 1 <> 0 then begin
    if (timer^.lap < timecount+clicktime*1000) then
    begin
      if @ondoubleclick <> nil then ondoubleclick(pos.x,pos.y,pressed and 1);
      event := event or evmousedouble;
    end else timecount := timer^.lap;
  end;

  if not drawing then begin
    lpos := pos;
    inc(pos.x,ev.relmove.x);
    inc(pos.y,ev.relmove.y);
    if pos.x < min.x then pos.x := min.x;
    if pos.x > max.x then pos.x := max.x;
    if pos.y < min.y then pos.y := min.y;
    if pos.y > max.y then pos.y := max.y;
  end;

  if @onmouseany <> nil then onmouseany(pos.x,pos.y,pressed);
  if (lpos.x <> pos.x)or(lpos.y <> pos.y) then begin
    if (visible)and(not frozen) then
      drawcursor(self,pos,screen^);
    if @onmousemove <> nil then onmousemove(pos.x,pos.y,pressed);
  end;
  if (up <> 0) then
    if @onmouseup <> nil then onmouseup(pos.x,pos.y,up);
  if (down <> 0) then
    if @onmousedown <> nil then onmousedown(pos.x,pos.y,down);
  if (screen^.mode.flags and mdlinear = 0) then
   if bank <> lastbank then setbank(bank);
end;

procedure sendmouseevent(mouse : word);
var
  a : pmhregistered;
begin
  a := dummy.next;
  while (a <> nil) do begin
    if a^.obj^.mouse = mouse then
    a^.obj^.handlemouseevent(globalevent[mouse]);
    a := a^.next;
  end;
end;


var
  moveint : boolean;

procedure mousehandle_drv;far;assembler;
asm
  mov   si,seg @DATA
  mov   ds,si
  db 66h;pusha
  cmp   moveint,true
  je    @end
  mov   moveint,true

  xor   bh,bh
  mov   globalevent[0].tintevent.buttons,bx

  and   ax,1
  jz    @nomove

    mov   ax,000Bh
    int   33h
    mov   globalevent[0].tintevent.relmove.x,cx
    mov   globalevent[0].tintevent.relmove.y,dx
    jmp   @moved

@nomove:
    mov   globalevent[0].tintevent.relmove.x,0
    mov   globalevent[0].tintevent.relmove.y,0
@moved:

  push  0
  call  sendmouseevent
@exit:
  mov   moveint,false

@end:
  db 66h;popa
end;

procedure mousehandle_com(inbyte,mouse : word);
var
  x,y   : integer;
const
  treshold = 6;
  ycount = 5;

begin
  dataqueue[mouse,queuepos[mouse]] := inbyte;
  if (dataqueue[mouse,queuepos[mouse]] = ord('M')) then begin
    queuepos[mouse] := 0;
    exit;
  end;
  inc(queuepos[mouse]);
  if (queuepos[mouse] = 3) then begin
    queuepos[mouse] := 0;
    if (dataqueue[mouse,0] and $40 = 0)or(moveint) then exit;
    moveint := true;
    globalevent[mouse].buttons := (dataqueue[mouse,0] and $20) shr 5+(dataqueue[mouse,0] and $10) shr 3;
    x := (dataqueue[mouse,1] and 63)+(dataqueue[mouse,0] and 3) shl 6;
    y := (dataqueue[mouse,2] and 63)+(dataqueue[mouse,0] and 12) shl 4;
    asm
      mov  ax,x
      shl  ax,8
      sar  ax,8
      mov  x,ax

      mov  ax,y
      shl  ax,8
      sar  ax,8
      mov  y,ax
    end;
    if (x <> 0) then begin
      if (abs(x) >= treshold) then inc(movecount[mouse].x)
      else if (movecount[mouse].x > 0) then dec(movecount[mouse].x);
      if (movecount[mouse].x >= ycount) then begin
        x := x*2;
        dec(movecount[mouse].x);
      end;
    end else if movecount[mouse].x > 0 then dec(movecount[mouse].x);
    if (y <> 0) then begin
      if (abs(y) >= treshold) then inc(movecount[mouse].y)
      else if (movecount[mouse].y > 0) then dec(movecount[mouse].y);
      if (movecount[mouse].y >= ycount) then begin
        y := y*2;
        dec(movecount[mouse].y);
      end;
    end else if movecount[mouse].y > 0 then dec(movecount[mouse].y);

    globalevent[mouse].relmove.x := x;
    globalevent[mouse].relmove.y := y;
    sendmouseevent(mouse);
    moveint := false;
  end;
end;

procedure mousehandle_com1;interrupt;
begin
asm
  db 66h;pusha
end;
  mousehandle_com(inp(com_baseaddr[1]+RxR),mouse_list[1]);
  outp($20,$20);
asm
  db 66h;popa
end;
end;

procedure mousehandle_com2;interrupt;
begin
asm
  db 66h;pusha
end;
  mousehandle_com(inp(com_baseaddr[2]+RxR),mouse_list[2]);
  outp($20,$20);
asm
  db 66h;popa
end;
end;

procedure mousehandle_com3;interrupt;
begin
asm
  db 66h;pusha
end;
  mousehandle_com(inp(com_baseaddr[3]+RxR),mouse_list[3]);
  outp($20,$20);
asm
  db 66h;popa
end;
end;

procedure mousehandle_com4;interrupt;
begin
asm
  db 66h;pusha
end;
  mousehandle_com(inp(com_baseaddr[4]+RxR),mouse_list[4]);
  outp($20,$20);
asm
  db 66h;popa
end;
end;

procedure com_openport(com : word;intproc: pointer);
begin
  getintvec(com_irqaddr[com]+8,oldint[com]);
  setintvec(com_irqaddr[com]+8,intproc);
  outp($21,inp($21) and not (1 shl com_irqaddr[com]));
  outp(com_baseaddr[com]+MCR,11);
  outp(com_baseaddr[com]+IER,1);
end;

procedure com_closeport(com : word);
begin
  outp(com_baseaddr[com]+MCR,0);
  outp(com_baseaddr[com]+IER,0);
  setintvec(com_irqaddr[com]+8,oldint[com]);
  outp($21,inp($21) or (1 shl com_irqaddr[com]));
end;

procedure com_setspeed(com : word;bps : longint);
var
  divisor : word;
begin
  outp(com_baseaddr[com]+LCR,inp(com_baseaddr[com]+LCR) or 128);
  divisor                     := 115200 div bps;
  outp(com_baseaddr[com]+DLL,lo(divisor));
  outp(com_baseaddr[com]+DLH,hi(divisor));
  outp(com_baseaddr[com]+LCR,inp(com_baseaddr[com]+LCR) and not 128);
end;

procedure com_setparam(com,data,parity,stop : word);
begin
  outp(com_baseaddr[com]+LCR,(data-5) or parity or ((stop-1) shl 2));
end;

var
  endtime     : longint;
  mouse_found : boolean;

function currenttime : longint;assembler;
asm
  mov es,[seg0040]
  mov ax,es:[6Ch]
  mov dx,es:[6Eh]
end;


procedure searchint_com;interrupt;
var
  b : byte;
begin
asm
  db 66h;pusha
end;
  b := inp(com_baseaddr[mouse_com[searchmouse]]+RxR);
  if b = ord('M') then mouse_found := true;
  outp($20,$20);
asm
  db 66h;popa
end;
end;

function searchmouse_com(mouse,com : word) : boolean;
var
  i : integer;
begin
  mouse_found := false;
  searchmouse := mouse;
  queuepos[mouse] := 0;
  com_openport(com,@searchint_com);
  com_setspeed(com,1200);
  com_setparam(com,7,N,1);
  endtime := currenttime+28;
  searchmouse_com := false;
  repeat
  until (mouse_found)or(currenttime >= endtime);
  searchmouse_com := mouse_found;
  com_closeport(com);
end;


function initmouse_com(mouse,com : word) : boolean;
begin
  queuepos[mouse] := 0;
  case com of
    1 : com_openport(com,@mousehandle_com1);
    2 : com_openport(com,@mousehandle_com2);
    3 : com_openport(com,@mousehandle_com3);
    4 : com_openport(com,@mousehandle_com4);
  end;
  com_setspeed(com,1200);
  com_setparam(com,7,N,1);
  endtime := currenttime+28;
  repeat
    if (queuepos[mouse] > 0) then if (dataqueue[mouse,queuepos[mouse]] = ord('M')) then
      endtime := currenttime;
  until (currenttime >= endtime);
  queuepos[mouse] := 0;
end;

procedure donemouse_com(mouse : word);
begin
  com_closeport(mouse);
end;

function globalinitmouse(mouse : word) : word;
var
  a,b : word;
  startcom : word;
label
  stop;
begin
  if (mouse = 0) then begin
    asm
      xor   ax,ax
      xor   bx,bx
      int   33h
      mov   a,ax
      mov   b,bx
    end;
    if (a = $FFFF) then begin
      case b of
        $0000 : avbuttons[mouse] := 3;
        $FFFF : avbuttons[mouse] := 2;
        else avbuttons[mouse] := b;
      end;
      asm
        { set the pointer to the mouse-handler }
        mov   ax,12
        mov   cx,0FFFFh
        mov   dx,offset cs:mousehandle_drv
        push  cs
        pop   es
        int   33h
        { reset mouseposition }
        mov   ax,000Bh
        int   33h
        { hide the mouse }
    {    mov   ax,0002h
        int   33h}
      end;
      mouse_type[0] := mt_drv;
      goto stop;
    end;
  end;

  globalinitmouse := errdevnotfound;

{  if not directmouse then begin
    asm
      mov   ax,0024h
      int   33h
      mov   t,ch
    end;
  end else t := 2;
  case t of
    1 : getmousetype := 'Bus-Mouse';
    2 : getmousetype := 'Serial Mouse';
    3 : getmousetype := 'InPort Mouse';
    4 : getmousetype := 'PS/2 Mouse';
    5 : getmousetype := 'HP Mouse';
  end;}

{  if mouse_type[0] = mt_com;}
  if (mouse+1 > numofports) then exit;
  mouse_type[mouse] := mt_com;
  mouse_list[mouse+1] := mouse;
  mouse_com[mouse] := mouse+1;
  if (not searchmouse_com(mouse,mouse+1)) then exit;

  initmouse_com(mouse,mouse_com[mouse]);
  avbuttons[mouse] := 2;
stop:
  inc(mice);
  globalinitmouse := errnoerror;
end;

procedure globaldonemouse;
var
  i : integer;
begin
  if mouse_type[0] = mt_drv then
  asm
    mov   ax,12
    xor   cx,cx
    xor   dx,dx
    mov   es,cx
    int   33h
  end;
  for i := 1 to mice do if (mouse_list[i] <> 0) then donemouse_com(i);
  mice := 0;
end;

procedure crstd;external;
{$L crstd.obj}
procedure crwstd;external;
{$L crwstd.obj}
procedure crhndstd;external;
{$L crhndstd.obj}



var
  oldexit : pointer;

procedure ml_exit;far;
begin
  if (mice <> 0) then globaldonemouse;
  exitproc := oldexit;
end;

var
  i : integer;

begin
  for i := 0 to 3 do com_baseaddr[i+1] := memw[seg0040:i shl 1];
  numofports := 0;
  while com_baseaddr[numofports+1] <> 0 do inc(numofports);

  check4fail(pointertoimage(@crstd,crdefault));
  check4fail(pointertoimage(@crwstd,crhourglass));
  check4fail(pointertoimage(@crhndstd,crhandpoint));
  oldexit  := exitproc;
  exitproc := @ml_exit;
end.


