unit tifflib;

interface

uses stream32,objects,newfrontier,gx_glob,gx_image,gx_col;

function info_tif(var stream : tstream32;var image_info : timageinfo) : word;
function loadtif(var stream : tstream32;var image : timage;usebpp,flags : longint) : word;

implementation

var
  maxread : word;

type
  tiffh = record
    numbertype             : word;
    version                : Word;
    offset                 : LongInt;
  End;

Const
     SubFileType = $FF;
     ImageWidth  = $100;
     ImageLength = $101;
     RowsPerStrip = $116;
     StripOffsets = $111;
     StripByteCounts= $117;
     SamplesPerPixel = $115;
     BitsPerSample   = $102;
     PlanarConfiguration = $11C;
     Compression = $103;
     Group3Options = $124;
     Group4Options = $125;
     FillOrder = $10A;
     Thresholding = $107;
     CellWidth = $108;
     MinSampleValue = $118;
     MaxSampleValue = $119;
     PhotometricInterp = $106;
     GrayResponseCurve = $123;
     ColorResponseUnit = $12C;
     ColorResponseCurves = $12D;
     XResolution = $11A;
     YResolution = $11B;
     ResolutionUnit = $128;
     Orientation = $112;
     DocumentName = $10D;
     PageName = $11D;
     XPosition = $11E;
     YPosition = $11F;
     PageNumber = $129;
     ImageDescription = $10E;
     Make = $10F;
     Model = $110;
     FreeOffsets = $120;
     FreeByteCounts = $121;
     PaletteOffset = $140;


     {Tiff Sizes}
     TIFFByte = 1;
     TIFFascii = 2;
     TIFFshort = 3;
     TIFFlong = 4;
     TIFFrational = 5;

     {TIFF compression types}
     COMPnone = 1;
     COMPhuff = 2;
     COMPfax3 = 3;
     COMPfax4 = 4;
     COMPLZW  = 5;

     COMPwrdl = $8003;
     COMPPackBits = $8005;
var
  tiffoffset: LongInt;
  TIFFEntries:Word;
  tiffwidth,tiffdepth:Word;
  tiffImageStart:LongInt;
  CompressionType: Word;
  TIFFbitsPerPixel:Byte;
  ColorTIFF:Boolean;


Function FGetWord(var stream : tstream32): Word;
Var
   N:Word;
Begin
     stream.read32(farptr(@N),sizeof(N));
     FGetWord:=N;
End;

Function FGetLong (var stream : tstream32):LongInt;
Var
   L:LongInt;
Begin
     stream.read32(farptr(@L),sizeof(L));
     FGetLong:=L;
End;



Procedure ReadTIFFTag(var stream : tstream32;Var Tag,TagType:Integer; Var Length,Offset:LongInt);
Begin
     tag:=fgetWord(stream);
     TagType:=fGetWord(stream);
     If TagType =TIFFLong then Begin
        length:=fgetLong(stream);
        Offset:=fgetLong(stream);
     End
    else
     Begin
          length:=fgetWord(stream);
          fgetword(stream);
          Offset:=fgetWord(stream);
          fgetword(stream);
     End;
     Case Tag of
          SubFileType:;
          ImageWidth: tiffWidth:=Offset;
          ImageLength: tiffDepth:=Offset;
        RowsPerStrip: ;
        StripOffsets: Begin
                           If TagType=TIFFLong then
                              TiffImageStart:=OffSet
                          else
                              TiffImageStart:=Offset AND $FFFF;
                           {is:=TiffImageStart;}
                      End;
        BitsPerSample:Begin
                          tiffbitsperPixel:=Offset;
                      End;
        Compression : Begin
                           CompressionType:=Offset;
                      End;

        PhotometricInterp:Begin
                               If (Offset=0) or (Offset=1) then
                                 ColorTIFF:=FALSE
                              else
                                 ColorTIFF:=TRUE;
                          End;

     End;

End;


function info_tif(var stream : tstream32;var image_info : timageinfo) : word;
var
  idvalid : boolean;
  i       : integer;
  tiffhead:tiffh;
  Tag,TagType:Integer;
  Length,Offset:LongInt;
  f:text;
begin
  info_tif := errfileerror;
  if (stream.status <> stok) then exit;
  stream.read32(farptr(@tiffhead),sizeof(tiffhead));
  if (stream.status <> stok) then Begin
     exit;
  End;
  idvalid := false;
  {check for II or MM}
  If (tiffhead.numbertype=18761) then
    idvalid := true;
  if (not idvalid) then begin
    info_tif := errfalseformat;
    stream.done;
    exit;
  end;
  {check the version}
  if tiffhead.version<>42 then begin
    info_tif := errfalseformat;
    stream.done;
    exit;
  end;
  TIFFOffset:=tiffhead.offset;
  stream.seek(tiffoffset);
  stream.read32(farptr(@tiffentries),sizeof(tiffentries));

  For I:=1 to TIFFEntries do
      ReadTIFFTag(stream, Tag,TagType,Length,Offset);

  with image_info do begin
    sizex        := tiffwidth;
    sizey        := tiffdepth;
    If ColorTIFF then
      bpp:=24
   else
      bpp:=8;
    if bpp <= 8 then inc(memsize,(1 shl bpp) shl 4);

    greyscale    := not (ColorTIFF);
    alphachannel := false;
    transparency := false;
    compressed   := CompressionType<>CompNone;
    case CompressionType of
      CompNone : comptype := 'None';
      ComPHuff : CompType := 'Huffman';
      CompFax3 : Comptype := 'CCITT FAX Group 3';
      CompFax4 : Comptype := 'CCITT FAX Group 4';
      else comptype := 'unknown';
    end;
    {debug}
    comptype:='None';

    interlaced   := false;
    frames       := 1;
    trans_color  := 0;
    memsize      := longint(sizex)*sizey;
    disksize     := stream.getsize;
    supported    := CompressionType = compnone;
    version      := tiffhead.version;
  end;
  info_tif := errnoerror;
end;

function loadtif(var stream : tstream32;var image : timage;usebpp,flags : longint) : word;
var
  y,mempos    : word;
  lineread    : word;
  readmax     : word;
  imgpos      : longint;
  cnvx        : word;
  idvalid     : boolean;
  i           : integer;
  convertdata : tcnvproc;
  cnvtmp      : pcnvproc;
  imageInfo   : timageinfo;
  Status      : Word;
  f:Text;
  p:Pointer;
begin
  loadtif := errfileerror;
  if (stream.status <> stok) then exit;
  status:=info_tif(stream,imageinfo);
  If status<>errnoerror then Begin
     loadtif:=status;
     stream.done;
     exit;
  End;

  if (imageinfo.comptype = 'unknown' ) or
      (not imageinfo.bpp in [1,4,8,24])
  then begin
    loadtif := errfalseformat;
    stream.done;
    exit;
  end;

  if (stream.errorinfo <> stok) then begin
    loadtif := errfalseend;
    stream.done;
    exit;
  end;

  cnvtmp  := prepareimage_cnv(image,imageinfo.sizex,imageinfo.sizey,
                 imageinfo.bpp,usebpp,0,flags,cnvimgmd or cnvswap);
  if (cnvtmp = nil) then begin
    loadtif := errfalsebpp;
    stream.done;
    exit;
  end;

  convertdata := tcnvproc(cnvtmp);

  if (imageinfo.bpp <= 8) then begin
    stream.read32(cnvpal,(word(1) shl image.bpp) shl 2);
    if image.bpp > 8 then begin
      case image.bpp of
        15 : pal_32to15(cnvpal,cnvpal,(word(1) shl image.bpp));
        16 : pal_32to16(cnvpal,cnvpal,(word(1) shl image.bpp));
      end;
    end
  end;
  {stream.seek(image.imagestart);}
  lineread := ((longint(Image.sizex)*image.bpp) shr 3{+3}){ and not 3};
  stream.reset;
  maxread := 64000;
  readmax := (maxread div lineread)*lineread;
  getmem(p,succ(readmax));
  mempos := readmax;
  {imgpos := longint(image.bpl)*pred(image.sizey);}
  imgpos := 0;

  {this part is unclear still. To find the first byte of the image data}
  {
  assign(f,'DEBUG');
  rewrite(f);
  writeln(f,is);
  Close(f);
  }

  {stream.seek(is);}
  if image.bpp>8 then
    stream.seek(1170)
 else
  stream.seek(TIFFImageStart+16);

  dec(image.sizey);
  {dec(imageinfo.sizey);}

  cnvxpos := 0;
  cnvx := image.sizex;
  if image.bpp = 24 then cnvx := cnvx*3;
  if image.bpp = image.bpp then cnvx := lineread;
  cnvx:=lineread;
  for y := 0 to image.sizey-1 do Begin     {downto 0 do begin}
    if mempos = readmax then begin
      stream.read32(farptr(p),readmax);
      mempos := 0;
    end;
    cnvypos := y;
    convertdata(ptr48(seg(p^),ofs(p^)+mempos),ptr48(image.data,image.imagestart+imgpos),cnvx);
    inc(imgpos,image.bpl);
    inc(mempos,lineread);
  end;

  stream.done;
  freemem(p,succ(readmax));
  loadtif := errnoerror;
end;

begin
end.